var canvas = null;
var canvasRect;

// line
var lastPos = null;
var line = null;
var stroke = null;
var lineset = new Array();
var currentPos = 0;

// event listeners
var buttonListeners = new Array();

// external command
var command = null;

// wait timer
var timerCanvas = null;
var waitTimer = null;
var waitCounter = 0;

// color palette
var paletteColor = { left:0, top:0, width:0, height:0, 
                     division:0, colors:[] };

// line width palette
var lineWidthPalette = { left:0, top:0, width:0, height:0,
                         centers:[], widths:[], radii:[] };

// line width palette

function tr( key ) {
    try {
        var val = localizedStrings[ key ];
        if ( val == undefined ) return key;
        return val;
    } catch ( e ) {
        debug( e );
        return key;
    }
}

function prefInt( key, defval ) {
    if ( window.widget ) {
        try {
            var val = widget.preferenceForKey( key );
            if ( val ) {
                return parseInt( val );
            }
        } catch ( e ) {
            debug( e );
        }
    }
    return defval;
}

var GifStatus = new function() {
    var tm = "" + (new Date()).getTime();
    this.id        = tm.substring( tm.length - 6, tm.length );
    this.delay     = prefInt( 'delay', 6 );
    this.frames    = prefInt( 'frames', 8 );
    this.vibration = prefInt( 'wobble', 4 );
    this.skip      = prefInt( 'sampling', 4 );
    this.lineWidth = prefInt( 'linewidth', 2 );
    this.transparency = prefInt( 'transparency', 10 );
    this.lineColor    = "rgb(0,0,0)";
    this.backgroundColor = "white";
    this.shadow    = { position:[ 2,5 ], color:[ 200, 200, 200 ], 
                       display:true };
    this.filename  = "/tmp/animgif" + this.id + ".gif";
    this.animating = false;

    if ( window.widget ) {
        this.shadow.display = ( prefInt( 'shadow', 1 ) == 1 );
    }

    this.toString = function() {
        try {
            return "Delay " + this.delay + ", Wobble " + this.vibration + ", Line " + this.lineWidth + ", Sampling " + this.sampling + ", Shadow " + this.shadow.display + ", Frames " + this.frames + ", Transparency " + this.transparency;
        } catch ( e ) {
            return e;
        }
    }
}

function debug( message ) {
    //document.getElementById( "debugDiv" ).innerText = message;
     document.getElementById( "debugDiv" ).appendChild( document.createTextNode( message ) );
     document.getElementById( "debugDiv" ).appendChild( document.createElement( 'br' ) );
}

function Line( color, width ) {
    debug( color );
//     if ( ( color[ 0 ] == 255 && color[ 1 ] == 255 && color[ 2 ] == 255 ) 
//          || color == null || color == 'white' ) {
    if ( color == "white" || color == "rgb(255,255,255)" ) {
        //debug( "ERASER!!!" );
        this.color = null;
        this.widthe = null;
        this.isActive = true;
    } else {
        this.color = color;
        this.width = width;
    }

    this.eraser = null;
    this.data  = new Array();
    this.box   = null;
    this.push  = _Line__push;
    this.draw  = _Line__draw;
    this.write = _Line__write;
    this.toString = _Line__toString;

    //
    this.isEraser = _Line__isEraser;
    this.isErased = _Line__isErased;
    this.erase    = _Line__erase;
    this.setBox   = _Line__setBox;
    this.removeEraser = _Line__removeEraser;
}

function _Line__push( x, y ) {
    var point = [ x, y ];
    this.data.push( point );
    this.box   = null;
}

function _Line__setBox() {
    if ( this.data.length == 0 ) {
        this.box = null;
        return;
    }
    var minx = this.data[ 0 ][ 0 ];
    var miny = this.data[ 0 ][ 1 ];
    var maxx = minx;
    var maxy = miny;
    for ( var i = 1; i < this.data.length; i++ ) {
        var x = this.data[ i ][ 0 ];
        var y = this.data[ i ][ 1 ];
        if ( x < minx ) minx = x;
        if ( x > maxx ) maxx = x;
        if ( y < miny ) miny = y;
        if ( y > maxy ) maxy = y;
    }
    this.box = [ minx, miny, maxx, maxy ];
}

function _Line__toString() {
    return "Line : " + this.width + " " + this.color + " " + this.data.length;
}

function _Line__isEraser() {
    return ( this.color == null );
}

function _Line__isErased() {
    return ( this.eraser != null && this.eraser.isActive );
}

function _Line__erase( eraser, x0, y0, x1, y1 ) {
    if ( this.isErased() || this.isEraser() ) return false;
    if ( this.box == null ) {
        this.setBox();
        if ( this.box == null ) return false;
    }
    var minx = this.box[ 0 ];
    var miny = this.box[ 1 ];
    var maxx = this.box[ 2 ];
    var maxy = this.box[ 3 ];
    if ( ( minx > x0 && minx > x1 ) || ( miny > y0 && miny > y1 )
         || ( maxx < x0 && maxx < x1 ) || ( maxy < y0 && maxy < y0 ) ) {
        return false;
    }
    var a = y0 - y1;
    var b = x1 - x0;
    var c = - a * x0 - b * y0;
    var s0 = a * this.data[ 0 ][ 0 ] + b * this.data[ 0 ][ 1 ] + c;
    //debug( "A = " + a + ", B = " + b + " C = " + c + " : s = " + s0 );
    for ( var i = 1; i < this.data.length; i++ ) {
        var s1 = a * this.data[ i ][ 0 ] + b * this.data[ i ][ 1 ] + c;
        //debug( "s0, s1 = " + s0 + ", " + s1 );
        if ( ( s0 < 0 && s1 >= 0 ) || ( s0 >= 0 && s1 < 0 ) ) {
            this.eraser = eraser;
            //debug( "detect cross" );
            return true;
        }
        s0 = s1;
    }
    return false;
}

function _Line__removeEraser( eraser ) {
    if ( this.eraser == eraser ) {
        this.eraser = null;
        return true;
    }
    return false;
}

function _Line__draw( context ) {
    var data = this.data;
    if ( data.length && data.length > 0 && 
         this.isEraser() == false && this.isErased() == false ) {
        context.lineWidth  = this.width;
        //context.fillStyle  = this.color;//"black";
        context.strokeStyle = this.color;//"black";
        context.moveTo( data[ 0 ][ 0 ], data[ 0 ][ 1 ] );
        for ( var i = 1; i < data.length; i++ ) {
            context.lineTo( data[ i ][ 0 ], data[ i ][ 1 ] );
        }
        context.stroke();
        context.save();
        debug( "COLOR " + this.color );
    }
}

function _Line__write( command, skip ) {
    if ( skip == undefined ) skip = 1;
    if ( this.isEraser() || this.isErased() ) return;
    var rgb = this.color.split( ',' );
    var r = 0;
    var g = 0;
    var b = 0;
    if ( rgb.length == 3 ) {
        var r = parseInt( rgb[ 0 ].substring( 4, rgb[ 0 ].length ) );
        var g = parseInt( rgb[ 1 ] );
        var b = parseInt( rgb[ 2 ].substring( 0, rgb[ 2 ].length - 1 ) );
    }
//    command.write( "c,0,0,0\n" );
    command.write( "c," + r + "," + g + "," + b + "\n" );
    command.write( "w," + this.width  + "\n");
    command.write( "l" );
    for ( var i = 0; i < this.data.length - 1; i+= skip ) {
        command.write( "," + this.data[ i ][ 0 ] + "," + this.data[ i ][ 1 ] );
    }
    command.write( "," + this.data[ this.data.length - 1 ][ 0 ] 
                   + "," + this.data[ this.data.length - 1 ][ 1 ] );
    command.write( "\n" );
}

function setup() {
    setCanvas( document.getElementById( "canvas" ) );
    if ( window.widget ) {
        widget.onhide   = hideWidget;
        widget.onremove = removeTemporaryFile; 
    }

    timerCanvas = document.getElementById( 'timer' );

    
    // buttons
    buttonListeners[ 'clear' ] 
        = setButtonAction( document.getElementById( 'clearButton' ), 
                           clearCanvas ); 
    buttonListeners[ 'stop' ] 
        = setButtonAction( document.getElementById( 'stopButton' ), 
                           stopMotion );
    buttonListeners[ 'animation' ] 
        = setButtonAction( document.getElementById( 'animButton' ), 
                           makeMotion );
    buttonListeners[ 'redo' ] 
        = setButtonAction( document.getElementById( 'redoButton' ), 
                         redoLine );
    buttonListeners[ 'undo' ] 
        = setButtonAction( document.getElementById( 'undoButton' ), undoLine )
;

    setCanvasTransparency();

    // menus
    //setLineWidthMenu();
    setWobbleMenu();
    setShadowMenu();
    setSamplingMenu();
    setDelayMenu();
    setTransparencyMenu();

    createGenericButton( document.getElementById( 'done' ),
                         tr( "Done" ), flipToFront );


    // shortcut
    window.onkeydown = keyPressed;

    // palette
    makePalette();
    makeLineWidthPalette();

    setButtonAvailability();
    //debug( GifStatus );

}

function setCanvasTransparency() {
    var canvas_bg = document.getElementById( "tracing-background" );
    if ( canvas_bg ) {
        canvas_bg.style.opacity = GifStatus.transparency * .1;
    }
}

function makeLineWidthPalette() {
    var palette = document.getElementById( "lwpalette" );
    if ( palette == null ) return;
    var ps = document.defaultView.getComputedStyle( palette );
    lineWidthPalette.top    = parseInt( ps.getPropertyValue( 'top' ) );
    lineWidthPalette.width  = parseInt( ps.getPropertyValue( 'width' ) );
    lineWidthPalette.height = parseInt( ps.getPropertyValue( 'height' ) );
    lineWidthPalette.left = ( window.outerWidth - parseInt( ps.getPropertyValue( "right" ) ) ) - lineWidthPalette.width;

    lineWidthPalette.widths  = [ 8, 6, 4, 2, 1 ];
    lineWidthPalette.centers = new Array();
    lineWidthPalette.radii   = new Array();

    var y = 2;
    for ( var i = 0; i < lineWidthPalette.widths.length; i++ ) {
        var r = lineWidthPalette.widths[ i ] * .75;
        lineWidthPalette.centers.push( y + r );
        lineWidthPalette.radii.push( r );
        y += r * 2 + 3;
    }

    drawLineWidthPalette();
}

function makePalette() {
    var colors = new Array();
    colors.push( [   0,   0,   0 ] ); // black
    colors.push( null ); // white = eraser
    //colors.push( [ 255, 255, 255 ] ); // white = eraser
    colors.push( [ 128, 128, 128 ] ); // gray

    colors.push( [ 255,   0,   0 ] ); // red
    colors.push( [   0, 255,   0 ] ); // lime
    colors.push( [   0,   0, 255 ] ); // blue

    colors.push( [ 255, 255,   0 ] ); // yellow
    colors.push( [ 255,   0, 255 ] ); // fuchsia
    colors.push( [   0, 255, 255 ] ); // aqua

    colors.push( [ 128,   0,   0 ] ); // maroon
    colors.push( [   0, 128,   0 ] ); // green
    colors.push( [   0,   0, 128 ] ); // navy

    colors.push( [ 128,   0, 128 ] ); // purple
    colors.push( [   0, 128, 128 ] ); // teal
    colors.push( [ 128, 128,   0 ] ); // olive

    //colors.push( [ 255, 192, 128 ] ); 
    colors.push( [  255, 128, 64 ] ); // orange

    paletteColor.colors = new Array();//colors;
    for ( var i = 0; i < colors.length; i++ ) {
        if ( colors[ i ] == null || colors[ i ].length < 3 ) {
            paletteColor.colors.push( "white" );
        } else {
            paletteColor.colors.push( "rgb(" + colors[ i ][ 0 ] + "," + colors[ i ][ 1 ] + "," + colors[ i ][ 2 ] + ")" );
        }
    }
    paletteColor.division = 4;

    var palette = document.getElementById( 'palette' );
    if ( palette == null ) return;

    var ps = document.defaultView.getComputedStyle( palette );
    paletteColor.division = 4;
    paletteColor.top    = parseInt( ps.getPropertyValue( 'top' ) );
    paletteColor.width  = parseInt( ps.getPropertyValue( 'width' ) );
    paletteColor.height = parseInt( ps.getPropertyValue( 'height' ) );
    paletteColor.left = ( window.outerWidth - parseInt( ps.getPropertyValue( "right" ) ) ) - paletteColor.width;

    drawPalette();
    //drawLineWidthPalette();
}

function drawLineWidthPalette() {
    var palette = document.getElementById( "lwpalette" );
    if ( palette == null || lineWidthPalette == null ) return;
    var context = palette.getContext( "2d" );
    context.clearRect( 0, 0, lineWidthPalette.width, lineWidthPalette.height );

    context.strokeStyle = "rgb(128,128,128)";
    context.lineWidth = 4.0;
    context.fillStyle ="black";
    var widths = lineWidthPalette.widths;
    var x = lineWidthPalette.width / 2;
//    var y = 2;
    debug( GifStatus.lineWidth + " : " +  lineWidthPalette.radii.length + ", " + x );
    debug( lineWidthPalette.left + ", " + lineWidthPalette.height + ", " + lineWidthPalette.top + ", " + lineWidthPalette.width );
//     lineWidthPalette.top    = parseInt( ps.getPropertyValue( 'top' ) );
//     lineWidthPalette.width  = parseInt( ps.getPropertyValue( 'width' ) );
//     lineWidthPalette.height = parseInt( ps.getPropertyValue( 'height' ) );
    for ( var i = 0; i < widths.length; i++ ) {
        var r = lineWidthPalette.radii[ i ];
        var y = lineWidthPalette.centers[ i ];
        //context.arc( x, y + r, r, 0, Math.PI * 2, 0 );
        if ( parseInt( GifStatus.lineWidth ) == lineWidthPalette.widths[ i ] ) {
            context.strokeStyle = "white";
            context.arc( x, y, r, 0, Math.PI * 2, 0 );
            context.stroke();
//         } else {
//             context.strokeStyle = "rgb(128,128,128)";
        }
//        context.arc( x, y + r, r, 0, Math.PI * 2, 0 );
        context.arc( x, y, r, 0, Math.PI * 2, 0 );
        context.fill();
        //context.save();
//        y += r * 2 + 3;
    }
    context.save();
}

function drawPalette() {
    var palette = document.getElementById( 'palette' );
    if ( palette == null ) return;
    var context = palette.getContext( "2d" );
    context.clearRect( 0, 0, paletteColor.width, paletteColor.height );
    var colors = paletteColor.colors;
    var div    = paletteColor.division;
    var dx     = paletteColor.width / div;
    var dy     = paletteColor.height / div;
    var lcol   = GifStatus.lineColor;
    debug( 'dx ' + dx + ", dy " + dy + ", lcol " + lcol + ", width " + paletteColor.width + ", height " + paletteColor.height + ",div " + div );
    var radius = dx * .5 - 1;
    if ( dx > dy ) radius = dy * .5 - 1;
    for ( var i = 0; i < colors.length; i++ ) {
        var x = i - parseInt( i / div ) * div;
        var y = parseInt( i / div );
        var colname = paletteColor.colors[ i ];
        context.fillStyle = colname;
        context.arc( x * dx + dx / 2, y * dy + dy / 2, radius, 0, 6.29, 0 );
        context.fill();
        if ( colname == lcol ) {
            context.strokeStyle = "white";
            context.lineWidth = 2.0;
            debug( "colname " + colname + ", lcol " + lcol );
        } else {
            context.strokeStyle = "black";
            context.lineWidth = 1.0;
        }
        context.arc( x * dx + dx / 2, y * dy + dy / 2, radius, 0, 6.29, 0 );
        context.stroke();
        context.save();
    }
    //context.save();
    //context.restore();

}

function changeLineWidth( event ) {
    if ( lineWidthPalette == null ) return;
    var x = event.x - lineWidthPalette.left;
    var y = event.y - lineWidthPalette.top;
    //debug( "( " + x + ", " + y + ")" );
    var selected = 0;
    var dist_min = Math.abs( lineWidthPalette.centers[ 0 ] - y );
    for ( var i = 1; i < lineWidthPalette.centers.length; i++ ) {
        var dist = Math.abs( lineWidthPalette.centers[ i ] - y );
        if ( dist <= dist_min ) {
            selected = i;
            dist_min = dist;
        }
    }
    GifStatus.lineWidth = lineWidthPalette.widths[ selected ];
    drawLineWidthPalette();
    if ( window.widget ) {
        widget.setPreferenceForKey( '' + parseInt( GifStatus.lineWidth ), 'linewidth' );
        //widget.setPreferenceForKey( 
    }
    debug( "selected " + selected + ", " + lineWidthPalette.widths[ selected ] );
}

function changeColor( event ) {
//     debug( event.x + ', ' + event.y + ", " + 
//            ( event.x - paletteColor.left ) + ", " 
//            + ( event.y - paletteColor.top ) );
    var x = event.x - paletteColor.left;
//     debug( "X = " + x );
    var y = event.y - paletteColor.top;
//     debug( "X = " + x );
    var div    = paletteColor.division;
//     debug( "div = " + div );
    var dx     = paletteColor.width / div;
//     debug( "dX = " + dx );
    var dy     = paletteColor.height / div;
//     debug( "dy = " + dy );
    var num = parseInt( y / dy ) * div + parseInt( x / dx );
//     debug( "num = " + num );
    GifStatus.lineColor = paletteColor.colors[ num ];
    drawPalette();
//     debug( "col = " + GifStatus.lineColor );
}

function keyPressed( event ) {
    debug( 'keypressed ' + event.keyCode );
    if ( event.shiftKey == false && event.altKey == false 
         && event.metaKey == true && event.ctrlKey == false 
         && event.keyCode == 90 ) {
        // Command + Z -> undo
        undoLine( event );
    } else if ( event.shiftKey == true && event.altKey == false 
                && event.metaKey == true && event.ctrlKey == false 
                && event.keyCode == 90 ) {
        // Shift + Command + Z -> redo 
        redoLine( event );
    } else if ( event.shiftKey == false && event.altKey == false 
                && event.metaKey == true && event.ctrlKey == false 
                && event.keyCode == 71 ) {
        // Shift + Command + G -> animation / stop
        if ( GifStatus.animating ) {
            stopMotion( event );
        } else {
            makeMotion( event );
        }
    }
}

function setButtonAvailability() {
    try {
        buttonListeners[ 'clear' ].enable( ( currentPos > 0 ) );
        buttonListeners[ 'stop' ].enable( GifStatus.animating );
        buttonListeners[ 'animation' ].enable( !(GifStatus.animating ) && currentPos > 0 );
        buttonListeners[ 'redo' ].enable( ( currentPos < lineset.length ) );
        buttonListeners[ 'undo' ].enable( ( currentPos > 0 ) );

    } catch ( e ) {
        debug( e );
    }
}

function setCanvas( canvasElement ) {
    canvas = canvasElement;
    var cstyle = document.defaultView.getComputedStyle( canvas );
    canvasRect = { left:parseInt( cstyle.getPropertyValue( 'left' ) ),
                   top:parseInt( cstyle.getPropertyValue( 'top' ) ),
                   width:parseInt( cstyle.getPropertyValue( 'width' ) ),
                   height:parseInt( cstyle.getPropertyValue( 'height' ) ) };
    var context = canvas.getContext( "2d" );
    //debug( canvasRect.width + ", " + canvasRect.height + " : " + context.fillStyle + ' , ' + context.lineWidth );
    context.fillStyle = "black";
    context.lineWidth = parseFloat( GifStatus.lineWidth );
    context.clearRect( 0, 0, canvasRect.width, canvasRect.height );
    context.lineCap = "round";
    context.lineJoin = "round";
    context.save();
}

function canvasMouseDown( event, targetCanvas ) {
    if ( targetCanvas == undefined ) {
        if ( canvas ) {
            targetCanvas = canvas;
        } else {
            setCanvas( document.getElementById( "canvas" ) );
            if ( targetCanvas == null ) {
                throw new Exception( "no canvas" );
            }
        }
    }
    if ( canvas != targetCanvas ) {
        setCanvas( targetCanvas );
    }

    var x = event.x - canvasRect.left;
    var y = event.y - canvasRect.top;
    lastPos = { x:x, y:y };

    var context = canvas.getContext( "2d" );
    context.lineWidth = GifStatus.lineWidth;
    context.strokeStyle = GifStatus.lineColor;
    var color = GifStatus.lineColor;
    stroke = new Line( color, parseFloat( GifStatus.lineWidth ) );

    window.addEventListener( 'mousemove', canvasMouseMove, true );
    window.addEventListener( 'mouseup', canvasMouseUp, true );
    event.stopPropagation();
    event.preventDefault();
}

function canvasMouseMove( event ) {
    var context = canvas.getContext( "2d" );
    var x = event.x - canvasRect.left;
    var y = event.y - canvasRect.top;
    if ( stroke.isEraser() == false ) {
        context.strokeStyle = stroke.color;
        context.moveTo( lastPos.x, lastPos.y );
        context.lineTo( x, y );
        context.stroke();
        context.save();
//     } else {
//         context.strokeStyle = "white";
    }
   
    // eraser
    if ( stroke.isEraser() ) {
        var repaintflag = false;
        for ( var i = 0; i < currentPos; i++ ) {
            if ( lineset[ i ].erase( stroke, x, y, lastPos.x, lastPos.y ) ) {
                repaintflag = true;
            }
        }
        if ( repaintflag ) {
            debug( "erased!!!!!!!" );
            redraw( lineset, 0, currentPos );
        }
    }

    lastPos = { x:x, y:y };
    stroke.push( x, y );
    //debug( stroke + " " + GifStatus.lineWidth);
}

function canvasMouseUp( event ) {
    if ( lineset.length > currentPos ) {
        lineset = lineset.slice( 0, currentPos );
    }

    if ( stroke.data.length > 1 ) {
        lineset.push( stroke );
        currentPos ++;
    }
    window.removeEventListener( 'mousemove', canvasMouseMove, true );
    window.removeEventListener( 'mouseup', canvasMouseUp, true );
    event.stopPropagation();
    event.preventDefault();
    setButtonAvailability();
}



function setLineWidth( width ) {
   var context = canvas.getContext( "2d" );
   context.lineWidth = width;
   context.save();
}

function redraw( lines, start, end ) {
    if ( lines == undefined ) lines = lineset;
    if ( start == undefined ) start = 0;
    if ( end == undefined )   end   = currentPos;

    var context = canvas.getContext( "2d" );

    try{
        context.clearRect( 0, 0, canvasRect.width, canvasRect.height );
        for ( var i = start; i < end; i++ ) {
            var line = lines[ i ];
            line.draw( context );
        }
        context.save();
        context.restore();
        //setButtonAvailability();
    } catch ( e ) {
        debug( e );
    }
}

function undoLine( event ) {
    debug( "undo" );
    if ( currentPos > 0 ) {
        currentPos --;
        if ( lineset[ currentPos ].isEraser() ) {
            lineset[ currentPos ].isActive = false;
        }
        redraw( lineset, 0, currentPos );
        debug( "undo finished" );
    } else {
        debug( "no more undo buffer" );
    }
    setButtonAvailability();
}

function redoLine( event ) {
    debug( "redo" );
    if ( GifStatus.animating ) {
        stopMotion();
    }
    if ( currentPos < lineset.length ) {
        if ( lineset[ currentPos ].isEraser() ) {
            lineset[ currentPos ].isActive = true;
        }
        //debug( "redo finished" );
        currentPos ++;
        redraw( lineset, 0, currentPos );
    } else {
        debug( "no more buffer" );
    }
    setButtonAvailability();    
}

function endHandler( com ) {
    var output = com.outputString;
    if ( output && output.length > 0 ) {
        var anim = document.getElementById( 'animationPanel' );
        var uri = com.outputString.split( "\n" )[ 0 ] + "?" + ( new Date() ).getTime();
        anim.src = uri;
        anim.style.display = "block";
        GifStatus.animating = true;
    } else {
        debug( "bad output '" + output + "': \n " + com.errorString );
    }
    if ( waitTimer ) {
        clearInterval( waitTimer );
        waitTimer = null;
        if ( timerCanvas ) {
            timerCanvas.style.display = "none";
        }
    }
    setButtonAvailability();
    command = null;
}

function stopMotion( event ) {
    var anim = document.getElementById( 'animationPanel' );
    anim.style.display = "none";
    GifStatus.animating = false;
    setButtonAvailability();
}

function clearCanvas( event ) {
    if ( GifStatus.animating ) {
        stopMotion();
    }
    if ( canvas ) {
        var context = canvas.getContext( "2d" );
        context.clearRect( 0, 0, canvasRect.width, canvasRect.height );
        context.save();
        lineset = new Array();
        currentPos = 0;
    }
    setButtonAvailability();
}

function hideWidget( event ) {
    stopMotion( event );
}

function removeTemporaryFile( event ) {
    if ( window.widget ) {
        // clear cache
        debug( widget.system( "/usr/bin/python cleanup.py " + GifStatus.filename, null ).errorString );
        //widget.system( "/bin/rm -f " + GifStatus.filename );
    }
}

function makeMotion( event ) {
    if ( command != null ) {
        command.cancel();
    }
    if ( waitTimer ) {
        clearInterval( waitTimer );
        waitTimer = null;
        timerCanvas.style.display = "none";
    }
    setTimeout( __makeMotion, 0 );
}

function __tickWait() {
    var tcontext = timerCanvas.getContext( "2d" );
    tcontext.clearRect( 0, 0, 16, 16 );
    var step = 3.14159265 * 2 / 12;
    var rad  = step * ( waitCounter % 12 );
    //tcontext.strokeStyle = "none";
    tcontext.strokeStyle = "none";
    for ( var i = 0; i < 4; i++ ) {
        var a = 0.2 * i + 0.4;
        var rgba = "rgba(128,32,32," + a + ")";
        tcontext.fillStyle = rgba;//"black";
        tcontext.arc( 8, 8, 8, rad, rad + step, 0 );
        tcontext.lineTo( 8, 8 );
        tcontext.fill();
        tcontext.save();
        rad += step;
    }
//     tcontext.arc( 8, 8, 8, rad, rad + 3.14159265 * 2 / 11, 0 );
//     tcontext.lineTo( 8, 8 );
    waitCounter ++;
    debug( waitCounter );
    //if ( waitCounter > 10 ) waitCounter = 0;
}

function __makeMotion() {
    debug( "motion" );
    if ( window.widget ) {
        // wait timer
        if ( timerCanvas ) {
            var tcontext = timerCanvas.getContext( "2d" );
            tcontext.clearRect( 0, 0, 20, 20 );
            waitCounter = 0;
            timerCanvas.style.display = "block";
            waitTimer = setInterval( __tickWait, 66 );
        }

        var arg_size   = " -s " + canvasRect.width + "x" + canvasRect.height;
        var arg_out    = " -o " + GifStatus.filename;

        var arg_shadow = "";
        if ( GifStatus.shadow && GifStatus.shadow.display ) {
            arg_shadow = " -shadow "; 
            var scol = GifStatus.shadow.color;
            for ( var i = 0; i < 3; i++ ) {
                if ( i != 0 ) arg_shadow += ",";
                arg_shadow += scol[ i ];
            }
            arg_shadow += " -spos " + GifStatus.shadow.position[ 0 ]
                + "," + GifStatus.shadow.position[ 1 ];
        }
        

        var arg_vibration = " -v " + GifStatus.vibration + "x" + GifStatus.vibration;
        var arg_frames = " -f " + GifStatus.frames;
        var arg_delay  = " -d " + GifStatus.delay;
        var pycommand  = "/usr/bin/python drawgif.py ";

        pycommand += arg_shadow;
        pycommand += arg_size;
        pycommand += arg_vibration;
        pycommand += arg_out;
        pycommand += arg_frames;
        pycommand += arg_delay;

        debug( pycommand );

        //
        command = widget.system(  pycommand, endHandler );
        try {
            for ( var i = 0; i < currentPos; i++ ) {//lineset.length; i++ ) {
                lineset[ i ].write( command, GifStatus.skip );
            }
        } catch ( e ) {
            debug( e );
        }
        command.write( "exit\n" );
        command.close();
    }
}


function imageDragStart( event ) {
    event.dataTransfer.effectAllowed = "copy";
    event.dataTransfer.setData( "URL", document.getElementById( 'animationPanel' ).src );
}

function flipToBehind( event ) {
    if ( window.widget ) {
        var front  = document.getElementById( "front" ); 
        var behind = document.getElementById( "behind" ); 
        widget.prepareForTransition("ToBack"); 
        behind.style.display  = "block"; 
        front.style.display = "none"; 
        setTimeout ('widget.performTransition();', 0);  
    }
}

function flipToFront( event ) {
    var front  = document.getElementById( "front" ); 
    var behind = document.getElementById( "behind" ); 
    if ( window.widget ) widget.prepareForTransition("ToBack"); 
    behind.style.display  = "none"; 
    front.style.display = "block"; 
    
    var values = new Array();
    //var items = [ 'linewidth', 'wobble', 'delay', 'sampling', 'shadow', 'transparency' ];
    var items = [ 'wobble', 'delay', 'sampling', 'shadow', 'transparency' ];
    for  ( var i = 0; i < items.length; i++ ) {        
        var value = parseInt( document.getElementById( items[ i ] ).value );
        values[ items[ i ] ] = value;
    }
    GifStatus.skip = values[ 'sampling' ];
    GifStatus.vibration = values[ 'wobble' ];
    GifStatus.delay = values[ 'delay' ];
    GifStatus.shadow.display = ( values[ 'shadow' ] == 1 );
    //GifStatus.lineWidth = values[ 'linewidth' ];
    GifStatus.transparency = values[ 'transparency' ];

    setCanvasTransparency();

    if ( window.widget ) {

        widget.setPreferenceForKey( '' + GifStatus.delay, 'delay' );
        //widget.setPreferenceForKey( '' + parseInt( GifStatus.lineWidth ), 'linewidth' );
        widget.setPreferenceForKey( '' + GifStatus.frames, 'frames' );
        widget.setPreferenceForKey( '' + GifStatus.vibration, 'wobble' );
        widget.setPreferenceForKey( '' + GifStatus.skip, 'sampling' );
        widget.setPreferenceForKey( '' + GifStatus.transparency, 'transparency' );
        if ( GifStatus.shadow.display ) {
            widget.setPreferenceForKey( '1' , 'shadow' );
        } else {
            widget.setPreferenceForKey( '0' , 'shadow' );
        }

        debug( "linewidth : " + '' + parseInt( GifStatus.lineWidth ) );

        // transit to front and repaint graphics
        setTimeout ('widget.performTransition();repaintAll();', 0);  
    }
}

function repaintAll() {
    redraw();
    drawPalette();
    drawLineWidthPalette();
    setButtonAvailability();
}


/// set preference panel
function setSelect( item ) {
    var label = document.getElementById( item.labelId );
    if ( label ) {
        label.innerText = tr( item.label );
    }
    var select = document.getElementById( item.selectId );
    select.innerHTML = "";
    for ( var i = 0; i < item.values.length; i++ ) {
        var option = document.createElement( 'option' );
        option.setAttribute( 'value', item.values[ i ] );
        option.appendChild( document.createTextNode( tr( item.labels[ i ] ) ) );
        if ( item.defaultValue == item.values[ i ] ) {
            option.setAttribute( 'selected' );
        }
        select.appendChild( option );
    }
}

function setLineWidthMenu() {
    var item = { labelId: "LineWidthLabel", label:"Line Width", 
                 selectId: "linewidth", 
                 defaultValue:GifStatus.lineWidth,
                 values: [ 8, 4, 2, 1 ],
                 labels: [ "Thicker", "Thick", "Middle", "Thin" ] };
    setSelect( item );
}

function setWobbleMenu() {
    var item = { labelId: "WobbleLabel", label:"Wobbling", 
                 selectId: "wobble", 
                 defaultValue:GifStatus.vibration,
                 values: [ 10, 7, 4, 3 ],
                 labels: [ "Larger", "Large", "Middle", "Small" ] };//, "Smaller" ] };
    setSelect( item );
}

function setShadowMenu() {
    var defval;
    if ( GifStatus.shadow.display ) defval = 1;
    else defval = 0;
    var item = { labelId: "ShadowLabel", label:"Shadow", 
                 selectId: "shadow", 
                 defaultValue:defval,
                 values: [ 1, 0 ],
                 labels: [ "With Shadow", "No Shadow" ] }
    setSelect( item );
}

function setSamplingMenu() {
    var item = { labelId: "SamplingLabel", label:"Sampling", 
                 selectId: "sampling", 
                 defaultValue:GifStatus.skip,
                 values: [ 32, 16, 8, 4, 2, 1 ],
                 labels: [ "Coasest", "Coaser", "Coase", "Middle", "Fine", "Finer" ] }
    setSelect( item );
}

function setDelayMenu() {
    var item = { labelId: "DelayLabel", label:"Delay", 
                 selectId: "delay", 
                 defaultValue:GifStatus.delay,
                 values: [ 100, 50, 20, 6, 3 ],
                 labels: [ "Slower", "Slow", "Middle", "Fast", "Faster" ] }
    setSelect( item );
}

function setTransparencyMenu() {
    var item = { labelId:"TransparencyLabel", label:"Canvas Transparency",
                 selectId: "transparency",
                 defaultValue:GifStatus.transparency,
                 values: [ 1, 5, 10 ],
                 labels: [ "Transparent", "Half", "Opaque" ] };
    setSelect( item );
}

// buttons

function ButtonListener( element, target ) {
    this.element = element;
    this.target = target;
    this.enabled   = true;

    this.mouseDown = _ButtonListener__mouseDown;
    this.mouseUp   = _ButtonListener__mouseUp;

    this.enable    = _ButtonListener__enable;

    var self = this;
    element.addEventListener( 'mousedown', function() { self.mouseDown(); } );
    element.addEventListener( 'mouseup', function() { self.mouseUp(); } );
    
}

function _ButtonListener__enable( flag ) {
    if ( flag == this.enabled ) return;
    this.enabled = flag;
    if ( flag ) {
        this.element.style.color = "#400";
    } else {
        this.element.style.color = "#c84";
    }
}

function _ButtonListener__mouseDown( event ) {
    if ( this.enabled == false ) return;
    this.element.style.background = "url(Images/Button2.png)";
}

function _ButtonListener__mouseUp( event ) {
    if ( this.enabled == false ) return;
    this.element.style.background = "url(Images/Button.png)";
    this.target( event );
}

function setButtonAction( element, targetFunc ) {
    var listener = new ButtonListener( element, targetFunc );
    return listener;
}
